from enum import Enum


#
# AMAAS exceptions class.
#
class AMaasException(Exception):
    def __init__(self, error_code, *params):
        self.error_code = error_code
        self.params = params
        self.message = error_code.value % params

    def __str__(self):
        return f"{self.error_code.name}: {self.message}"


#
# Error codes for AMAAS exceptions
#
class AMaasErrorCode(Enum):
    MSG_ID_ERR_FILE_NOT_FOUND = "Failed to open file. No such file or directory %s."
    MSG_ID_ERR_FILE_NO_PERMISSION = "Failed to open file. Permission denied to open %s."
    MSG_ID_ERR_INVALID_REGION = "%s is not a supported region, region value should be one of %s"
    MSG_ID_ERR_MISSING_AUTH = "Must provide an API key to use the client."
    MSG_ID_GRPC_ERROR = "Received gRPC status code: %s, msg: %s."
    MSG_ID_ERR_KEY_AUTH_FAILED = "Invalid token or Api Key."
    MSG_ID_ERR_UNKNOWN_CMD = "Received unknown command from server: %d"
    MSG_ID_ERR_UNKNOWN_STAGE = "Received unknown stage from server: %d"
    MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE = "Received unexpected command %d and stage %d."
    MSG_ID_ERR_UNEXPECTED_ERROR = "Unexpected error encountered. %s"
    MSG_ID_ERR_RATE_LIMIT_EXCEEDED = "Raised by the SDK library to indicate http 429 too many request error."
    MSG_ID_ERR_INVALID_TAG = "Invalid tag format: %s."
    MSG_ID_ERR_TAG_NUMBER_EXCEED = "Too many tags: %d."



import io
import os
from typing import BinaryIO, List

import grpc
import logging

from ..protos import scan_pb2
from ..protos import scan_pb2_grpc
from ..exception import AMaasException
from ..exception import AMaasErrorCode
from ..util import _init_by_region_util
from ..util import _init_util
from ..util import _validate_tags
from ..util import _digest_hex
from ..util import APP_NAME_HEADER, APP_NAME_FILE_SCAN

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO')
logger.setLevel(LOG_LEVEL)
logger.propagate = False

timeout_in_seconds = 180


def init_by_region(region, api_key, enable_tls=True, ca_cert=None):
    return _init_by_region_util(region, api_key, enable_tls, ca_cert, True)


def init(host, api_key=None, enable_tls=False, ca_cert=None):
    return _init_util(host, api_key, enable_tls, ca_cert, True)


async def quit(handle):
    await handle.close()


# https://github.com/grpc/grpc/blob/91083659fa88c938779dd41e57a7f97981b6c9a1/src/python/grpcio_tests/tests_aio/unit/channel_test.py#L180


async def _scan_data(channel: grpc.Channel, data_reader: BinaryIO, size: int, identifier: str, tags: List[str],
                     app_name: str) -> str:
    _validate_tags(tags)
    stub = scan_pb2_grpc.ScanStub(channel)
    stats = {}
    result = None
    try:
        metadata = (
            (APP_NAME_HEADER, app_name),
        )
        call = stub.Run(timeout=timeout_in_seconds, metadata=metadata)

        request = scan_pb2.C2S(stage=scan_pb2.STAGE_INIT,
                               file_name=identifier,
                               rs_size=size,
                               offset=0,
                               chunk=None,
                               tags=tags,
                               file_sha1="sha1:" + _digest_hex(data_reader, "sha1"),
                               file_sha256="sha256:" + _digest_hex(data_reader, "sha256"))

        await call.write(request)

        while True:
            response = await call.read()

            if response.cmd == scan_pb2.CMD_RETR:
                if response.stage != scan_pb2.STAGE_RUN:
                    raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE, response.cmd,
                                         response.stage)
                logger.debug(
                    f"stage RUN, try to read {response.length} at offset {response.offset}")

                data_reader.seek(response.offset)
                chunk = data_reader.read(response.length)

                request = scan_pb2.C2S(
                    stage=scan_pb2.STAGE_RUN,
                    file_name=None,
                    rs_size=0,
                    offset=data_reader.tell(),
                    chunk=chunk)

                stats["total_upload"] = stats.get(
                    "total_upload", 0) + len(chunk)
                await call.write(request)
            elif response.cmd == scan_pb2.CMD_QUIT:
                if response.stage != scan_pb2.STAGE_FINI:
                    raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE, response.cmd,
                                         response.stage)
                result = response.result
                logger.debug("receive QUIT, exit loop...\n")
                break
            else:
                logger.debug("unknown command...")
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNKNOWN_CMD, response.cmd)

        await call.done_writing()

        total_upload = stats.get("total_upload", 0)
        logger.debug(f"total upload {total_upload} bytes")

    except AMaasException:
        raise
    except grpc.aio.AioRpcError as rpc_error:
        if "429" in str(rpc_error):
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_RATE_LIMIT_EXCEEDED)
        elif rpc_error.code() == grpc.StatusCode.UNAUTHENTICATED:
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_KEY_AUTH_FAILED)
        else:
            raise AMaasException(AMaasErrorCode.MSG_ID_GRPC_ERROR, rpc_error.code().value[0], rpc_error.details())
    except Exception as err:
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_ERROR, str(err))

    return result


async def scan_file(channel: grpc.Channel, file_name: str, tags: List[str] = None,
                    app_name: str = APP_NAME_FILE_SCAN) -> str:
    try:
        f = open(file_name, "rb")
        fid = os.path.basename(file_name)
        n = os.stat(file_name).st_size
    except FileNotFoundError as err:
        logger.debug("File not exist: " + str(err))
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_FILE_NOT_FOUND, file_name)
    except (PermissionError, IOError) as err:
        logger.debug("Permission error: " + str(err))
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_FILE_NO_PERMISSION, file_name)
    return await _scan_data(channel, f, n, fid, tags, app_name)


async def scan_buffer(channel: grpc.Channel, bytes_buffer: bytes, uid: str, tags: List[str] = None,
                      app_name: str = APP_NAME_FILE_SCAN) -> str:
    f = io.BytesIO(bytes_buffer)
    return await _scan_data(channel, f, len(bytes_buffer), uid, tags, app_name)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Stage(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    STAGE_INIT: _ClassVar[Stage]
    STAGE_RUN: _ClassVar[Stage]
    STAGE_FINI: _ClassVar[Stage]

class Command(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    CMD_RETR: _ClassVar[Command]
    CMD_QUIT: _ClassVar[Command]
STAGE_INIT: Stage
STAGE_RUN: Stage
STAGE_FINI: Stage
CMD_RETR: Command
CMD_QUIT: Command

class C2S(_message.Message):
    __slots__ = ["stage", "file_name", "rs_size", "offset", "chunk", "trendx", "file_sha1", "file_sha256", "tags"]
    STAGE_FIELD_NUMBER: _ClassVar[int]
    FILE_NAME_FIELD_NUMBER: _ClassVar[int]
    RS_SIZE_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    CHUNK_FIELD_NUMBER: _ClassVar[int]
    TRENDX_FIELD_NUMBER: _ClassVar[int]
    FILE_SHA1_FIELD_NUMBER: _ClassVar[int]
    FILE_SHA256_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    stage: Stage
    file_name: str
    rs_size: int
    offset: int
    chunk: bytes
    trendx: bool
    file_sha1: str
    file_sha256: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, stage: _Optional[_Union[Stage, str]] = ..., file_name: _Optional[str] = ..., rs_size: _Optional[int] = ..., offset: _Optional[int] = ..., chunk: _Optional[bytes] = ..., trendx: bool = ..., file_sha1: _Optional[str] = ..., file_sha256: _Optional[str] = ..., tags: _Optional[_Iterable[str]] = ...) -> None: ...

class S2C(_message.Message):
    __slots__ = ["stage", "cmd", "offset", "length", "result"]
    STAGE_FIELD_NUMBER: _ClassVar[int]
    CMD_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    LENGTH_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    stage: Stage
    cmd: Command
    offset: int
    length: int
    result: str
    def __init__(self, stage: _Optional[_Union[Stage, str]] = ..., cmd: _Optional[_Union[Command, str]] = ..., offset: _Optional[int] = ..., length: _Optional[int] = ..., result: _Optional[str] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: amaas/grpc/protos/scan.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x61maas/grpc/protos/scan.proto\x12\ramaas.scan.v1\"\xb3\x01\n\x03\x43\x32S\x12#\n\x05stage\x18\x01 \x01(\x0e\x32\x14.amaas.scan.v1.Stage\x12\x11\n\tfile_name\x18\x02 \x01(\t\x12\x0f\n\x07rs_size\x18\x03 \x01(\x05\x12\x0e\n\x06offset\x18\x04 \x01(\x05\x12\r\n\x05\x63hunk\x18\x05 \x01(\x0c\x12\x0e\n\x06trendx\x18\x06 \x01(\x08\x12\x11\n\tfile_sha1\x18\x07 \x01(\t\x12\x13\n\x0b\x66ile_sha256\x18\x08 \x01(\t\x12\x0c\n\x04tags\x18\t \x03(\t\"\x7f\n\x03S2C\x12#\n\x05stage\x18\x01 \x01(\x0e\x32\x14.amaas.scan.v1.Stage\x12#\n\x03\x63md\x18\x02 \x01(\x0e\x32\x16.amaas.scan.v1.Command\x12\x0e\n\x06offset\x18\x03 \x01(\x05\x12\x0e\n\x06length\x18\x04 \x01(\x05\x12\x0e\n\x06result\x18\x05 \x01(\t*6\n\x05Stage\x12\x0e\n\nSTAGE_INIT\x10\x00\x12\r\n\tSTAGE_RUN\x10\x01\x12\x0e\n\nSTAGE_FINI\x10\x02*%\n\x07\x43ommand\x12\x0c\n\x08\x43MD_RETR\x10\x00\x12\x0c\n\x08\x43MD_QUIT\x10\x01\x32;\n\x04Scan\x12\x33\n\x03Run\x12\x12.amaas.scan.v1.C2S\x1a\x12.amaas.scan.v1.S2C\"\x00(\x01\x30\x01\x42\x33\n\x1d\x63om.trend.cloudone.amaas.scanZ\x12\x61maas/scanner/baseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'amaas.grpc.protos.scan_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.trend.cloudone.amaas.scanZ\022amaas/scanner/base'
  _globals['_STAGE']._serialized_start=358
  _globals['_STAGE']._serialized_end=412
  _globals['_COMMAND']._serialized_start=414
  _globals['_COMMAND']._serialized_end=451
  _globals['_C2S']._serialized_start=48
  _globals['_C2S']._serialized_end=227
  _globals['_S2C']._serialized_start=229
  _globals['_S2C']._serialized_end=356
  _globals['_SCAN']._serialized_start=453
  _globals['_SCAN']._serialized_end=512
# @@protoc_insertion_point(module_scope)



import re
from os.path import abspath, dirname, join
from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

base_dir = abspath(dirname(__file__))
package_name = "amaas"

with open(join(base_dir, package_name, "_version.py")) as f:
    package_version = re.search(
        r"__version__\s+=\s+[\"\']([^\"\']+)[\"\']", f.read()
    ).group(1)

setup(install_requires=['grpcio==1.56.2', 'protobuf==4.23.4'],
      name="visionone-filesecurity",
      version=package_version,
      author="Trend Micro VisionOne File Security Team",
      description="Trend Micro VisionOne File Security SDK for python",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/trendmicro/tm-v1-fs-python-sdk",
      packages=find_namespace_packages(exclude=['tests*', 'examples']),
      package_data={'amaas': ['grpc/protos/*']},
      include_package_data=True,
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          "License :: OSI Approved :: MIT License",
],
    python_requires='>=3.7',
)



import grpc
import hashlib
from typing import BinaryIO, List
from .exception import AMaasException
from .exception import AMaasErrorCode

HASH_CHUNK_SIZE = 512 * 1024

APP_NAME_HEADER = "tm-app-name"
APP_NAME_FILE_SCAN = "V1FS"

# regions
AWS_JP_REGION = "ap-northeast-1"
AWS_SG_REGION = "ap-southeast-1"
AWS_AU_REGION = "ap-southeast-2"
AWS_IN_REGION = "ap-south-1"
AWS_US_REGION = "us-east-1"
AWS_DE_REGION = "eu-central-1"
AWS_CA_REGION = "ca-central-1"
AWS_TREND_REGION = "us-east-2"
AWS_GB_REGION = "eu-west-2"
C1_JP_REGION = "jp-1"
C1_SG_REGION = "sg-1"
C1_AU_REGION = "au-1"
C1_IN_REGION = "in-1"
C1_US_REGION = "us-1"
C1_DE_REGION = "de-1"
C1_CA_REGION = "ca-1"
C1_TREND_REGION = "trend-us-1"
C1_GB_REGION = "gb-1"

C1Regions = [C1_AU_REGION, C1_CA_REGION, C1_DE_REGION, C1_GB_REGION, C1_IN_REGION, C1_JP_REGION, C1_SG_REGION,
             C1_US_REGION, C1_TREND_REGION]
V1Regions = [AWS_AU_REGION, AWS_DE_REGION, AWS_IN_REGION, AWS_JP_REGION, AWS_SG_REGION, AWS_US_REGION]
SupportedV1Regions = [AWS_AU_REGION, AWS_DE_REGION, AWS_JP_REGION, AWS_SG_REGION, AWS_US_REGION]
SupportedC1Regions = [C1_AU_REGION, C1_CA_REGION, C1_DE_REGION, C1_GB_REGION, C1_IN_REGION, C1_JP_REGION, C1_SG_REGION,
                      C1_US_REGION]

AllRegions = C1Regions + V1Regions
AllValidRegions = SupportedC1Regions + SupportedV1Regions

V1ToC1RegionMapping = {AWS_AU_REGION: C1_AU_REGION,
                       AWS_DE_REGION: C1_DE_REGION,
                       AWS_IN_REGION: C1_IN_REGION,
                       AWS_JP_REGION: C1_JP_REGION,
                       AWS_SG_REGION: C1_SG_REGION,
                       AWS_US_REGION: C1_US_REGION,
                       }


class _GrpcAuth(grpc.AuthMetadataPlugin):
    def __init__(self, key):
        self._key = key

    def __call__(self, context, callback):
        callback((('authorization', self._key),), None)


def _init_util(host, api_key=None, enable_tls=False, ca_cert=None, is_aio_channel=False):
    call_creds = None
    if api_key:
        auth_key_str = 'ApiKey ' + api_key
        call_creds = grpc.metadata_call_credentials(_GrpcAuth(auth_key_str))

    if enable_tls:
        if ca_cert:
            # Bring Your Own Certificate case
            with open(ca_cert, 'rb') as f:
                ssl_creds = grpc.ssl_channel_credentials(f.read())
        else:
            ssl_creds = grpc.ssl_channel_credentials()

        # if authentication is necessary, combined call_creds with ssl_creds.
        # Otherwise, just use ssl_creds for channel credentials.
        if call_creds is None:
            creds = ssl_creds
        else:
            creds = grpc.composite_channel_credentials(ssl_creds, call_creds)

        channel = grpc.aio.secure_channel(host, creds) if is_aio_channel else grpc.secure_channel(host, creds)
    else:
        channel = grpc.aio.insecure_channel(host) if is_aio_channel else grpc.insecure_channel(host)

    return channel


def _init_by_region_util(region, api_key, enable_tls=True, ca_cert=None, is_aio_channel=False):
    mapping = {
        C1_US_REGION: 'antimalware.us-1.cloudone.trendmicro.com:443',
        C1_IN_REGION: 'antimalware.in-1.cloudone.trendmicro.com:443',
        C1_DE_REGION: 'antimalware.de-1.cloudone.trendmicro.com:443',
        C1_SG_REGION: 'antimalware.sg-1.cloudone.trendmicro.com:443',
        C1_AU_REGION: 'antimalware.au-1.cloudone.trendmicro.com:443',
        C1_JP_REGION: 'antimalware.jp-1.cloudone.trendmicro.com:443',
        C1_GB_REGION: 'antimalware.gb-1.cloudone.trendmicro.com:443',
        C1_CA_REGION: 'antimalware.ca-1.cloudone.trendmicro.com:443',
    }

    # make sure it is valid V1 or C1 region
    if region not in SupportedV1Regions:
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_INVALID_REGION, region, SupportedV1Regions)
    else:
        # map it to C1 region if it is V1 region
        c1_region = V1ToC1RegionMapping.get(region)
        if not c1_region:
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_INVALID_REGION, region, SupportedV1Regions)
        region = c1_region

    host = mapping.get(region, None)
    if host is None:
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_INVALID_REGION, region)
    return _init_util(host, api_key, enable_tls, ca_cert, is_aio_channel)


def _validate_tags(tags: List[str]):
    if tags is not None:
        if len(tags) > 8:
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_TAG_NUMBER_EXCEED, len(tags))

        for t in tags:
            if not (0 < len(t) < 64):
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_INVALID_TAG, t)


def _digest_hex(data_reader: BinaryIO, algorithm: str):
    if algorithm == "sha1":
        file_hash = hashlib.sha1()
    elif algorithm == "sha256":
        file_hash = hashlib.sha256()
    else:
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_ERROR, "unsupported hash algorithm " + algorithm)

    w = data_reader.tell()
    data_reader.seek(0)

    chunk = data_reader.read(HASH_CHUNK_SIZE)
    while chunk:
        file_hash.update(chunk)
        chunk = data_reader.read(HASH_CHUNK_SIZE)

    data_reader.seek(w)
    return file_hash.hexdigest()
